// 2009 Hiroyuki Ogasawara
// vim:ts=4 sw=4 noet:


#include	<stdio.h>
#include	<linux/input.h>
#include	<sys/ioctl.h>
#include	<string.h>
#include	"linux/uinput.h"
#include	"hook_main.h"

#define	EXTERN_C	

EXTERN_C const char*	name= "ctrl-a";
EXTERN_C const char*	version_information= "ctrl-a v1.10\n";
EXTERN_C const char*	description= "ctrl-a v1.10 (http://wlog.flatlib.jp)\n";


#define	__USE_DELAY_TIME	0
#define	DELAY_AKEY_TIME		100

#define	__USE_IGNORE_AKEY	1
#define	IGNORE_AKEY_TIME	1000

enum {
	FALSE	=	0,
	TRUE	=	1,
};

enum {
	VALUE_KEYUP,
	VALUE_KEYDOWN,
	VALUE_KEYREPEAT,
};

enum {
	DFLAG_A_DOWN		=	(1<<0),
	DFLAG_CTRL			=	(1<<1),
	DFLAG_DELAY_TIME	=	(1<<2),
};

enum {
	ST_INIT,
	ST_CTRL_DOWN,
	ST_A_DOWN,
	ST_KEY_DOWN,
};

enum {
	MAX_KEY_STATE	=	KEY_MAX,
	MAX_PUSH_BUFFER	=	16,
};

typedef struct _T_STATE {
	int	State;
	int	Counter;
} T_STATE;


static int		DownFlag= 0;
static int		pushBuffer[MAX_PUSH_BUFFER];
static int		pushCounter= 0;
static int		keyboard_fd= 0;
static int		A_Counter= 0;
static int		delay_Time= 0;
static int		delay_Key= 0;
static int		downa_Time= 0;
static T_STATE	KeyState[MAX_KEY_STATE];

static void
Init()
{
	DownFlag= 0;
	memset( KeyState, 0, sizeof(T_STATE) * MAX_KEY_STATE );
	memset( pushBuffer, 0, sizeof(int) * MAX_PUSH_BUFFER );
	pushCounter= 0;
	keyboard_fd= 0;
}

static void
SetFlag( int flag )
{
	DownFlag|= flag;
}

static void
ResetFlag( int flag )
{
	DownFlag&= ~flag;
}

static int
TestFlag( int flag )
{
	return	DownFlag & flag;
}


EXTERN_C void*
init_event_handler( int number_of_event_device_files, int* event_fds )
{
	Init();
	keyboard_fd= event_fds[0];
	ioctl( keyboard_fd, EVIOCGRAB, 1 );
	return	NULL;
}


EXTERN_C void
finalize_event_hander( void* user_data )
{
	ioctl( keyboard_fd, EVIOCGRAB, 0 );
}


static void
FlushKey()
{
	int	i;
	for( i= 0 ; i< pushCounter ; i++ ){
		int	vk= pushBuffer[ i ];
		send_event( EV_KEY, vk, VALUE_KEYDOWN );
		KeyState[ vk ].State= ST_KEY_DOWN;
	}
	pushCounter= 0;
}

static void
PushKey( int vk )
{
	if( pushCounter < MAX_PUSH_BUFFER ){
		pushBuffer[pushCounter++]= vk;
	}
}


static void
RunState( unsigned int ktime, int type, int vk, int value )
{
	T_STATE*	tp= &KeyState[vk];
	switch( tp->State ){
	case ST_INIT:
		if( value == VALUE_KEYDOWN ){
			if( TestFlag( DFLAG_CTRL ) ){
				send_event( type, vk, VALUE_KEYDOWN );
				tp->State= ST_CTRL_DOWN;
			}else if( TestFlag( DFLAG_A_DOWN ) ){
				PushKey( vk );
				tp->Counter= A_Counter;
				tp->State= ST_A_DOWN;
			}else{
				send_event( type, vk, VALUE_KEYDOWN );
				tp->State= ST_KEY_DOWN;
			}
		}else{
			send_event( type, vk, value );
		}
		break;
	case ST_A_DOWN:
		if( TestFlag( DFLAG_A_DOWN ) && tp->Counter == A_Counter ){
			if( __USE_DELAY_TIME && value == VALUE_KEYUP ){
				if( !TestFlag( DFLAG_CTRL|DFLAG_DELAY_TIME ) ){
					delay_Time= ktime;
					delay_Key= vk;
					tp->State= ST_INIT;
					SetFlag( DFLAG_DELAY_TIME );
					break;
				}
			}
			if( !TestFlag( DFLAG_CTRL ) ){
				send_event( type, KEY_LEFTCTRL, VALUE_KEYDOWN );
				SetFlag( DFLAG_CTRL );
			}
			FlushKey();
		}
		send_event( type, vk, value );
		if( value == VALUE_KEYUP ){
			tp->State= ST_INIT;
		}else{
			tp->State= ST_KEY_DOWN;
		}
		break;
	case ST_CTRL_DOWN:
	case ST_KEY_DOWN:
		send_event( type, vk, value );
		if( value == VALUE_KEYUP ){
			tp->State= ST_INIT;
		}
		break;
	}
}


EXTERN_C void
handle_event( int fd, struct input_event event, void* user_data )
{
	int	vk= event.code;
	int	type= event.type;
	int	value= event.value;
#if __USE_DELAY_TIME || __USE_IGNORE_AKEY
	unsigned int	ktime= event.time.tv_sec * 1000 + event.time.tv_usec / 1000;
#else
	unsigned int	ktime= 0;
#endif

	switch( type ){
	case EV_REL:
		send_event( type, vk, value );
		break;

	case EV_KEY:
		switch( vk ){
		case KEY_A:
			if( value == VALUE_KEYDOWN ){
				A_Counter++;
				ResetFlag( DFLAG_CTRL );
				SetFlag( DFLAG_A_DOWN );
				downa_Time= ktime;
			}else if( value == VALUE_KEYREPEAT ){
				// nop
			}else if( value == VALUE_KEYUP ){
				A_Counter++;
				if( TestFlag( DFLAG_A_DOWN ) ){
					if( __USE_DELAY_TIME && TestFlag( DFLAG_DELAY_TIME ) ){
						if( ktime - delay_Time < DELAY_AKEY_TIME ){
							send_event( EV_KEY, KEY_A, VALUE_KEYDOWN );
							send_event( EV_KEY, KEY_A, VALUE_KEYUP );
							FlushKey();
						}else{
							send_event( type, KEY_LEFTCTRL, VALUE_KEYDOWN );
							SetFlag( DFLAG_CTRL );
							FlushKey();
							send_event( type, delay_Key, VALUE_KEYUP );
							send_event( type, KEY_LEFTCTRL, VALUE_KEYUP );
						}
						ResetFlag( DFLAG_DELAY_TIME );
					}else if( TestFlag( DFLAG_CTRL ) ){
						send_event( type, KEY_LEFTCTRL, VALUE_KEYUP );
					}else{
						if( ktime - downa_Time < IGNORE_AKEY_TIME ){
							send_event( EV_KEY, KEY_A, VALUE_KEYDOWN );
							send_event( EV_KEY, KEY_A, VALUE_KEYUP );
						}
						FlushKey();
					}
				}
				ResetFlag( DFLAG_A_DOWN|DFLAG_CTRL );
			}
			break;
		default:
			RunState( ktime, type, vk, value );
			break;
		}
		break;

	case EV_SYN:
		send_event( type, vk, value );
		break;

	default:
		break;
	}
}



EXTERN_C void
set_event_bits( int ui_fd )
{
	int	i;

	ioctl( ui_fd, UI_SET_EVBIT, EV_KEY );
	ioctl( ui_fd, UI_SET_EVBIT, EV_SYN );

	for( i= 0 ; i< MAX_KEY_STATE ; i++ ){
		ioctl( ui_fd, UI_SET_KEYBIT, i );
	}
}


